@@ -1796,15 +1796,23 @@ static fsal_status_t ceph_fsal_commit2(struct fsal_obj_handle *obj_hdl,
 	bool closefd = false;
 	struct ceph_export *export =
 		container_of(op_ctx->fsal_export, struct ceph_export, export);
+	struct user_cred root_creds = {};
+	struct user_cred *saved_creds = op_ctx->creds;
 
-	/* Make sure file is open in appropriate mode.
-	 * Do not check share reservation.
+
+	/*
+	 * Make sure file is open in appropriate mode, without checking for
+	 * share reservation. Also, it's possible that the file has changed
+	 * permissions since it was opened by the writer, so open the file
+	 * with root creds here since we're just doing a fsync.
 	 */
+	op_ctx->creds = &root_creds;
 	status = fsal_reopen_obj(obj_hdl, false, false, FSAL_O_WRITE,
 				 (struct fsal_fd *)&myself->fd, &myself->share,
 				 ceph_open_func, ceph_close_func,
 				 (struct fsal_fd **)&out_fd, &has_lock,
 				 &closefd);
+	op_ctx->creds = saved_creds;
 
 	if (!FSAL_IS_ERROR(status)) {
 		retval = ceph_ll_fsync(export->cmount, out_fd->fd, false);
@@ -2207,6 +2215,9 @@ static fsal_status_t ceph_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 		struct timespec timestamp;
 
 		mask |= CEPH_SETATTR_ATIME;
+	#ifdef CEPH_SETATTR_ATIME_NOW
+		mask |= CEPH_SETATTR_ATIME_NOW;
+	#endif
 		rc = clock_gettime(CLOCK_REALTIME, &timestamp);
 		if (rc != 0) {
 			LogDebug(COMPONENT_FSAL,
@@ -2227,6 +2238,9 @@ static fsal_status_t ceph_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 		struct timespec timestamp;
 
 		mask |= CEPH_SETATTR_MTIME;
+	#ifdef CEPH_SETATTR_MTIME_NOW
+		mask |= CEPH_SETATTR_MTIME_NOW;
+	#endif
 		rc = clock_gettime(CLOCK_REALTIME, &timestamp);
 		if (rc != 0) {
 			LogDebug(COMPONENT_FSAL,
