@@ -151,7 +151,7 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 	}
 
 	construct_handle(glfs_export, &sb, glhandle, globjhdl,
-			 GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
+			 &objhandle, vol_uuid);
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes_all(&sb, attrs_out);
@@ -387,7 +387,7 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 	}
 
 	construct_handle(glfs_export, &sb, glhandle, globjhdl,
-			 GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
+			 &objhandle, vol_uuid);
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes_all(&sb, attrs_out);
@@ -513,7 +513,7 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 	}
 
 	construct_handle(glfs_export, &sb, glhandle, globjhdl,
-			 GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
+			 &objhandle, vol_uuid);
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes_all(&sb, attrs_out);
@@ -614,7 +614,7 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 	}
 
 	construct_handle(glfs_export, &sb, glhandle, globjhdl,
-			 GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
+			 &objhandle, vol_uuid);
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes_all(&sb, attrs_out);
@@ -1631,7 +1631,7 @@ static fsal_status_t glusterfs_open2(struct fsal_obj_handle *obj_hdl,
 	}
 
 	construct_handle(glfs_export, &sb, glhandle, globjhdl,
-			 GLAPI_HANDLE_LENGTH, &myself, vol_uuid);
+			 &myself, vol_uuid);
 
 	*new_obj = &myself->handle;
 
@@ -2052,6 +2052,7 @@ static fsal_status_t seek2(struct fsal_obj_handle *obj_hdl,
 	fsal_openflags_t openflags = FSAL_O_READ;
 	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
 	struct glusterfs_fd my_fd = {0};
+	struct stat sbuf = {0};
 	struct glusterfs_export *glfs_export =
 	   container_of(op_ctx->fsal_export, struct glusterfs_export, export);
 
@@ -2068,6 +2069,22 @@ static fsal_status_t seek2(struct fsal_obj_handle *obj_hdl,
 	if (FSAL_IS_ERROR(status))
 		goto out;
 
+	ret = glfs_fstat(my_fd.glfd, &sbuf);
+	if (ret != 0) {
+		if (errno == EBADF)
+			errno = ESTALE;
+		status = gluster2fsal_error(errno);
+		goto out;
+	}
+
+	/* RFC7862 15.11.3,
+	 * If the sa_offset is beyond the end of the file,
+	 * then SEEK MUST return NFS4ERR_NXIO. */
+	if (offset > sbuf.st_size) {
+		status = gluster2fsal_error(ENXIO);
+		goto out;
+	}
+
 	SET_GLUSTER_CREDS(glfs_export, &op_ctx->creds->caller_uid,
 			&op_ctx->creds->caller_gid,
 			op_ctx->creds->caller_glen,
@@ -2095,7 +2112,7 @@ static fsal_status_t seek2(struct fsal_obj_handle *obj_hdl,
 		}
 		goto out;
 	} else {
-		info->io_eof = FALSE;
+		info->io_eof = (ret >= sbuf.st_size) ? TRUE : FALSE;
 		info->io_content.hole.di_offset = ret;
 	}
 
@@ -2152,8 +2169,11 @@ static fsal_status_t glusterfs_commit2(struct fsal_obj_handle *obj_hdl,
 				  &op_ctx->creds->caller_gid,
 				  op_ctx->creds->caller_glen,
 				  op_ctx->creds->caller_garray);
-
+#ifdef USE_GLUSTER_STAT_FETCH_API
+		retval = glfs_fsync(out_fd->glfd, NULL, NULL);
+#else
 		retval = glfs_fsync(out_fd->glfd);
+#endif
 
 		if (retval == -1) {
 			retval = errno;
@@ -2465,9 +2485,12 @@ static fsal_status_t glusterfs_setattr2(struct fsal_obj_handle *obj_hdl,
 				  &op_ctx->creds->caller_gid,
 				  op_ctx->creds->caller_glen,
 				  op_ctx->creds->caller_garray);
-
+#ifdef USE_GLUSTER_STAT_FETCH_API
+		retval = glfs_ftruncate(my_fd.glfd, attrib_set->filesize,
+					NULL, NULL);
+#else
 		retval = glfs_ftruncate(my_fd.glfd, attrib_set->filesize);
-
+#endif
 		SET_GLUSTER_CREDS(glfs_export, NULL, NULL, 0, NULL);
 
 		if (retval != 0) {
