@@ -26,11 +26,35 @@
 #include "fsal_internal.h"
 #include "fsal_convert.h"
 #include "gpfs_methods.h"
+#include "nfs_init.h"
 #include <sys/types.h>
 #include <unistd.h>
 #include <utime.h>
 #include <sys/time.h>
 
+/* Setup up_vector. File system's upvector_mutex must be held */
+static bool setup_up_vector(struct gpfs_filesystem *gpfs_fs)
+{
+	struct gpfs_filesystem_export_map *map;
+
+	map = glist_first_entry(&gpfs_fs->exports,
+				struct gpfs_filesystem_export_map, on_exports);
+	if (!map)
+		return false;
+
+	gpfs_fs->up_vector = (struct fsal_up_vector *)map->exp->export.up_ops;
+
+	/* wait for upcall readiness */
+	up_ready_wait(gpfs_fs->up_vector);
+
+	/* Set up op_ctx for the thread */
+	op_ctx = &gpfs_fs->req_ctx;
+	op_ctx->fsal_export = gpfs_fs->up_vector->up_fsal_export;
+	op_ctx->ctx_export = gpfs_fs->up_vector->up_gsh_export;
+
+	return true;
+}
+
 /**
  * @brief Up Thread
  *
@@ -40,7 +64,7 @@
 void *GPFSFSAL_UP_Thread(void *Arg)
 {
 	struct gpfs_filesystem *gpfs_fs = Arg;
-	const struct fsal_up_vector *event_func;
+	struct fsal_up_vector *event_func;
 	char thr_name[16];
 	int rc = 0;
 	struct pnfs_deviceid devid;
@@ -54,7 +78,7 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 	int retry = 0;
 	struct gsh_buffdesc key;
 	uint32_t expire_time_attr = 0;
-	uint32_t upflags = 0;
+	uint32_t upflags;
 	int errsv = 0;
 	fsal_status_t fsal_status = {0,};
 
@@ -70,20 +94,15 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 		 gpfs_fs->fs->dev.major, gpfs_fs->fs->dev.minor);
 	SetNameFunction(thr_name);
 
-	/* Set the FSAL UP functions that will be used to process events. */
-	event_func = gpfs_fs->up_ops;
-
-	if (event_func == NULL) {
-		LogFatal(COMPONENT_FSAL_UP,
-			 "FSAL up vector does not exist. Can not continue.");
-		gsh_free(Arg);
-		return NULL;
-	}
-
 	LogFullDebug(COMPONENT_FSAL_UP,
 		     "Initializing FSAL Callback context for %d.",
 		     gpfs_fs->root_fd);
 
+	/* wait for nfs init completion to get general_fridge
+	 * initialized which is needed for processing some upcall events
+	 */
+	nfs_init_wait();
+
 	/* Start querying for events and processing. */
 	while (1) {
 		LogFullDebug(COMPONENT_FSAL_UP,
@@ -110,11 +129,6 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 		errsv = errno;
 
 		if (rc != 0) {
-			LogCrit(COMPONENT_FSAL_UP,
-				"OPENHANDLE_INODE_UPDATE failed for %d. rc %d, errno %d (%s) reason %d",
-				gpfs_fs->root_fd, rc, errsv,
-				strerror(errsv), reason);
-
 			rc = -(rc);
 			if (rc > GPFS_INTERFACE_VERSION) {
 				LogFatal(COMPONENT_FSAL_UP,
@@ -122,12 +136,20 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 					 callback.interface_version, rc);
 				return NULL;
 			}
-			if (retry < 1000) {
-				retry++;
-				continue;
-			}
 
-			if (errsv == EUNATCH)
+			if (errsv == EINTR)
+				continue;
+
+			LogCrit(COMPONENT_FSAL_UP,
+				"OPENHANDLE_INODE_UPDATE failed for %d. rc %d, errno %d (%s) reason %d",
+				gpfs_fs->root_fd, rc, errsv,
+				strerror(errsv), reason);
+
+			/* @todo 1000 retry logic will go away once the
+			 * OPENHANDLE_INODE_UPDATE ioctl separates EINTR
+			 * and EUNATCH.
+			 */
+			if (errsv == EUNATCH && ++retry > 1000)
 				LogFatal(COMPONENT_FSAL_UP,
 					 "GPFS file system %d has gone away.",
 					 gpfs_fs->root_fd);
@@ -176,6 +198,17 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 		LogDebug(COMPONENT_FSAL_UP, "Received event to process for %d",
 			 gpfs_fs->root_fd);
 
+		/* We need valid up_vector while processing some of the
+		 * events below. Setup up vector and hold the mutex while
+		 * processing the event for the entire duration.
+		 */
+		PTHREAD_MUTEX_lock(&gpfs_fs->upvector_mutex);
+		if (!setup_up_vector(gpfs_fs)) {
+			PTHREAD_MUTEX_unlock(&gpfs_fs->upvector_mutex);
+			return NULL;
+		}
+		event_func = gpfs_fs->up_vector;
+
 		switch (reason) {
 		case INODE_LOCK_GRANTED:	/* Lock Event */
 		case INODE_LOCK_AGAIN:	/* Lock Event */
@@ -199,14 +232,14 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 				if (reason == INODE_LOCK_AGAIN)
 					fsal_status = up_async_lock_avail(
 							 general_fridge,
-							 event_func->up_export,
+							 event_func,
 							 &key,
 							 fl.lock_owner,
 							 &lockdesc, NULL, NULL);
 				else
 					fsal_status = up_async_lock_grant(
 							 general_fridge,
-							 event_func->up_export,
+							 event_func,
 							 &key,
 							 fl.lock_owner,
 							 &lockdesc, NULL, NULL);
@@ -218,7 +251,7 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 				 "delegation recall: flags:%x ino %" PRId64,
 				 flags, callback.buf->st_ino);
 			fsal_status = up_async_delegrecall(general_fridge,
-						  event_func->up_export,
+						  event_func,
 						  &key, NULL, NULL);
 			break;
 
@@ -235,7 +268,7 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 
 				fsal_status = up_async_layoutrecall(
 							general_fridge,
-							event_func->up_export,
+							event_func,
 							&key,
 							LAYOUT4_NFSV4_1_FILES,
 							false, &segment,
@@ -273,7 +306,7 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 			devid.fsal_id = FSAL_ID_GPFS;
 
 			fsal_status = up_async_notify_device(general_fridge,
-						event_func->up_export,
+						event_func,
 						NOTIFY_DEVICEID4_DELETE_MASK,
 						LAYOUT4_NFSV4_1_FILES,
 						&devid,
@@ -313,7 +346,7 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 				 */
 				if (flags & (UP_SIZE | UP_SIZE_BIG)) {
 					fsal_status = event_func->invalidate(
-						event_func->up_export, &key,
+						event_func, &key,
 						FSAL_UP_INVALIDATE_CACHE);
 					break;
 				}
@@ -324,21 +357,15 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 				    ~(UP_SIZE | UP_NLINK | UP_MODE | UP_OWN |
 				     UP_TIMES | UP_ATIME | UP_SIZE_BIG)) {
 					fsal_status = event_func->invalidate(
-						event_func->up_export, &key,
+						event_func, &key,
 						FSAL_UP_INVALIDATE_CACHE);
 				} else {
 					/* buf may not have all attributes set.
-					 * Since posix2fsal_attributes() copies
-					 * all attributes and also sets
-					 * attr.mask, correct attr.mask with
-					 * valid upcall flags only before
-					 * passing the attr to update() which
-					 * actually updates the cache_inode
-					 * object attributes.
+					 * Set the mask to what is changed
 					 */
-					posix2fsal_attributes(&buf, &attr);
-					/* Set the mask to what is changed */
 					attr.valid_mask = 0;
+					attr.acl = NULL;
+					upflags = 0;
 					if (flags & UP_SIZE)
 						attr.valid_mask |=
 						   ATTR_CHGTIME | ATTR_CHANGE |
@@ -358,7 +385,8 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 					if (flags & UP_OWN)
 						attr.valid_mask |=
 						   ATTR_CHGTIME | ATTR_CHANGE |
-						   ATTR_OWNER;
+						   ATTR_OWNER | ATTR_GROUP |
+						   ATTR_MODE;
 					if (flags & UP_TIMES)
 						attr.valid_mask |=
 						   ATTR_CHGTIME | ATTR_CHANGE |
@@ -368,22 +396,27 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 						attr.valid_mask |=
 						   ATTR_CHGTIME | ATTR_CHANGE |
 						   ATTR_ATIME;
+					if (flags & UP_NLINK)
+						attr.valid_mask |=
+							ATTR_NUMLINKS;
+					attr.request_mask = attr.valid_mask;
 
 					attr.expire_time_attr =
 					    expire_time_attr;
 
-					fsal_status = event_func->
-					    update(event_func->up_export,
-						   &key, &attr, upflags);
+					posix2fsal_attributes(&buf, &attr);
+					fsal_status = event_func->update(
+							event_func, &key,
+							&attr, upflags);
 
 					if ((flags & UP_NLINK)
 					    && (attr.numlinks == 0)) {
-						upflags = fsal_up_nlink |
-							  fsal_up_close;
+						upflags = fsal_up_nlink;
 						attr.valid_mask = 0;
+						attr.request_mask = 0;
 						fsal_status = up_async_update
 						    (general_fridge,
-						     event_func->up_export,
+						     event_func,
 						     &key, &attr,
 						     upflags, NULL, NULL);
 					}
@@ -395,6 +428,7 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 			LogDebug(COMPONENT_FSAL_UP,
 				"Terminating the GPFS up call thread for %d",
 				gpfs_fs->root_fd);
+			PTHREAD_MUTEX_unlock(&gpfs_fs->upvector_mutex);
 			return NULL;
 
 		case INODE_INVALIDATE:
@@ -404,7 +438,7 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 
 			upflags = FSAL_UP_INVALIDATE_CACHE;
 			fsal_status = event_func->invalidate_close(
-						event_func->up_export,
+						event_func,
 						&key,
 						upflags);
 			break;
@@ -415,13 +449,17 @@ void *GPFSFSAL_UP_Thread(void *Arg)
 			 * eventually get other errors that stop this
 			 * thread.
 			 */
+			PTHREAD_MUTEX_unlock(&gpfs_fs->upvector_mutex);
 			continue; /* get next event */
 
 		default:
+			PTHREAD_MUTEX_unlock(&gpfs_fs->upvector_mutex);
 			LogWarn(COMPONENT_FSAL_UP, "Unknown event: %d", reason);
 			continue;
 		}
 
+		PTHREAD_MUTEX_unlock(&gpfs_fs->upvector_mutex);
+
 		if (FSAL_IS_ERROR(fsal_status) &&
 		    fsal_status.major != ERR_FSAL_NOENT) {
 			LogWarn(COMPONENT_FSAL_UP,
