@@ -43,6 +43,7 @@
 #include "nfs_exports.h"
 #include "pnfs_utils.h"
 #include "conf_url.h"
+#include "sal_functions.h"
 
 /**
  * @brief LTTng trace enabling magic
@@ -66,6 +67,7 @@
 #include "gsh_lttng/logger.h"
 #include "gsh_lttng/mdcache.h"
 #include "gsh_lttng/nfs_rpc.h"
+#include "gsh_lttng/nfs4.h"
 #include "gsh_lttng/state.h"
 #include "gsh_lttng/fsal_mem.h"
 #endif /* USE_LTTNG */
@@ -231,6 +233,23 @@ int nfs_libmain(const char *ganesha_conf,
 		goto fatal_die;
 	}
 
+
+	/* Create stable storage directory, this needs to be done before
+	 * starting the recovery thread.
+	 */
+	rc = nfs4_recovery_init();
+	if (rc) {
+		LogCrit(COMPONENT_INIT,
+			  "Recovery backend initialization failed!");
+		goto fatal_die;
+	}
+
+	/* Start grace period */
+	nfs_start_grace(NULL);
+
+	/* Wait for enforcement to begin */
+	nfs_wait_for_grace_enforcement();
+
 	/* Load export entries from parsed file
 	 * returns the number of export entries.
 	 */
