@@ -127,7 +127,7 @@ int nfs4_op_putrootfh(struct nfs_argop4 *op, compound_data_t *data,
 	set_current_entry(data, file_obj);
 
 	/* Put our ref */
-	file_obj->obj_ops.put_ref(file_obj);
+	file_obj->obj_ops->put_ref(file_obj);
 
 	/* Convert it to a file handle */
 	if (!nfs4_FSALToFhandle(data->currentFH.nfs_fh4_val == NULL,
