@@ -133,7 +133,6 @@ typedef struct nfs41_session_slot__ {
 	struct COMPOUND4res_extended cached_result;	/*< NFv41: pointer to
 							   cached RPC result in
 							   a session's slot */
-	unsigned int cache_used;	/*< If we cached the result */
 } nfs41_session_slot_t;
 
 /**
@@ -153,6 +152,14 @@ enum {
 	session_bc_fault = 0x02, /* not actually used anywhere */
 };
 
+/**
+ * @brief minimum values for channel attributes
+ */
+#define NFS41_MIN_REQUEST_SIZE 256
+#define NFS41_MIN_RESPONSE_SIZE 256
+#define NFS41_MIN_OPERATIONS 2
+#define NFS41_MAX_CONNECTIONS 16
+
 /**
  * @brief Structure representing an NFSv4.1 session
  */
@@ -174,7 +181,10 @@ struct nfs41_session {
 				   and on which we signal when we
 				   free an entry. */
 
-	SVCXPRT *xprt;		/*< Referenced pointer to transport */
+	pthread_rwlock_t conn_lock;
+	int num_conn;
+	sockaddr_t connections[NFS41_MAX_CONNECTIONS];
+
 	nfs_client_id_t *clientid_record;	/*< Client record
 						   correspinding to ID */
 	clientid4 clientid;	/*< Clientid owning this session */
@@ -253,13 +263,17 @@ struct state_share {
  */
 
 struct state_nlm_share {
-	struct glist_head share_perclient;	/*< Lock states for this
-						    open state
-						    This field MUST be first */
-	unsigned int share_access;	/*< The NFSv4 Share Access state */
-	unsigned int share_deny;	/*< The NFSv4 Share Deny state */
-	unsigned int share_access_prev;	/*< Previous share access state */
-	unsigned int share_deny_prev;	/*< Previous share deny state   */
+	/** Put this share state on a list per client. This field MUST be first
+	 */
+	struct glist_head share_perclient;
+	/** The NLM Share Access state */
+	unsigned int share_access;
+	/** The NLM Share Deny state */
+	unsigned int share_deny;
+	/** Counts of each share access state */
+	unsigned int share_access_counts[fsa_RW + 1];
+	/** Counts of each share deny state   */
+	unsigned int share_deny_counts[fsm_DRW + 1];
 };
 
 /**
